# encoding: utf-8
# vi:si:et:sw=4:sts=4:ts=4
from __future__ import division, with_statement, print_function, absolute_import

import os
import json
import shutil
import time
try:
    import _thread as thread
    from queue import Queue
except:
    import thread
    from Queue import Queue
from threading import Thread

import ox
from twisted.web.resource import Resource
from twisted.web.static import File
from twisted.web.server import Site
from twisted.internet import reactor

from . import extract
from .utils import hash_prefix

class UploadThread(Thread):
    def __init__(self, server):
        Thread.__init__(self)
        self.server = server

    def run(self):
        while True:
            oshash = self.server.upload.get()
            print(oshash)
            try:
                self.server.client.upload([oshash])
            except:
                print('failed to upload', oshash)
            self.server.upload.task_done()

class Server(Resource):

    def __init__(self, client):
        self.upload = Queue()
        self.client = client
        Resource.__init__(self)
        t = UploadThread(self)
        t.setDaemon(True)
        t.start()

    def active_encodes(self):
        conn, c = self.client._conn()
        site = self.client._config['url']
        active = int(time.mktime(time.localtime())) - 120
        status = 'active'
        sql = 'SELECT oshash FROM encode WHERE site = ? AND status = ? AND modified > ?'
        args = [site, status, active]
        c.execute(sql, tuple(args))
        files = [row[0] for row in c]
        # reset inactive encodes
        sql = 'UPDATE encode SET status = ? WHERE site = ? AND status = ? AND modified < ?'
        c.execute(sql, ('', site, 'active', active))
        conn.commit()
        conn.close()
        return files

    def is_available(self, oshash):
        info = self.client.info(oshash)
        if info and 'error' not in info:
            for path in self.client.path(oshash):
                if os.path.exists(path):
                    return True
        return False

    def queued(self):
        site = self.client._config['url']
        files = self.client.get_encodes(site)
        available = list(filter(lambda oshash: self.is_available(oshash), files))
        unavailable = list(set(files) - set(available))
        return available, unavailable

    def queued_encodes(self):
        available, unavailable = self.queued()
        return available

    def queue_available_uploads(self):
        files = self.queued_encodes()
        for oshash in files:
            info = self.client.info(oshash)
            if not info or 'error' in info:
                continue
            path = self.media_path(oshash)
            if os.path.exists(path) and os.stat(path).st_size > 0:
                self.update_status(oshash, 'done')
                self.upload.put(oshash)

    def update_status(self, oshash, status):
        conn, c = self.client._conn()
        site = self.client._config['url']
        modified = int(time.mktime(time.localtime()))
        c.execute(u'UPDATE encode SET status = ?, modified = ? WHERE site = ? AND oshash = ?', (status, modified, site, oshash))
        conn.commit()
        conn.close()

    def media_path(self, oshash):
        return os.path.join(
            self.client.media_cache(),
            os.path.join(*hash_prefix(oshash)),
            self.client.profile(self.client.info(oshash))
        )

    def render_json(self, request, response):
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(response, indent=2).encode()

    def getChild(self, name, request):
        # make source media available via oshash
        if request.path.startswith(b'/get/'):
            oshash = request.path.decode().split('/')[-1]
            for path in self.client.path(oshash):
                if os.path.exists(path):
                    f = File(path, 'application/octet-stream')
                    f.isLeaf = True
                    return f
        return self

    def render_PUT(self, request):
        if request.path.startswith(b'/upload'):
            parts = request.path.decode().split('/')
            oshash = parts[-1]
            if len(oshash) == 16:
                path = self.media_path(oshash)
                tmp = path + '.tmp.webm'
                ox.makedirs(os.path.dirname(path))
                with open(tmp, 'wb') as f:
                    shutil.copyfileobj(request.content, f)
                shutil.move(tmp, path)
                self.update_status(oshash, 'done')
                self.upload.put(oshash)
                return self.render_json(request, {
                    'path': path
                })
        request.setResponseCode(404)
        return '404 unkown location'

    def render_POST(self, request):
        if request.path.startswith(b'/status'):
            oshash = request.path.decode().split('/')[-1]
            self.update_status(oshash, 'failed')
            return self.render_json(request, {})
        request.setResponseCode(404)
        return '404 unkown location'

    def render_GET(self, request):
        if request.path.startswith(b'/next'):
            response = {}
            files = self.queued_encodes()
            for oshash in files:
                info = self.client.info(oshash)
                if not info or 'error' in info:
                    continue
                path = self.media_path(oshash)
                if os.path.exists(path) and os.stat(path).st_size > 0:
                    self.update_status(oshash, 'done')
                    self.upload.put(oshash)
                    continue
                for f in self.client.path(oshash):
                    if os.path.exists(f):
                        response['oshash'] = oshash
                        url = 'http://%s:%s/get/%s' % (request.host.host, request.host.port, oshash)
                        output = '/tmp/%s.%s' % (oshash, self.client.profile(info))
                        response['cmd'] = extract.video_cmd(url, output, self.client.profile(info), info)
                        if isinstance(response['cmd'][0], list):
                            for part in response['cmd']:
                                part[0] = 'ffmpeg'
                        else:
                            response['cmd'][0] = 'ffmpeg'
                        response['output'] = output
                        self.update_status(oshash, 'active')
                        print(oshash, f)
                        return self.render_json(request, response)
            return self.render_json(request, response)
        elif request.path.startswith(b'/ping/'):
            parts = request.path.decode().split('/')
            # FIXME: store client id somewhere
            client = parts[-1]
            oshash = parts[-2]
            self.update_status(oshash, 'active')
            return self.render_json(request, {})
        elif request.path.startswith(b'/update'):
            thread.start_new_thread(self.update, ())
            return self.render_json(request, {'status': True})
        elif request.path.startswith(b'/status'):
            queued, offline = self.queued()
            return self.render_json(request, {
                'active': self.active_encodes(),
                'queue': queued,
                'offline': offline
            })
        request.setHeader('Content-Type', 'text/html')
        data = b'pandora_client distributed encoding server'
        return data

    def update(self):
        self.client.scan([])
        self.client.sync([])
        self.client.update_encodes(True)

def run(client, args=None):
    if not args:
        args = []
    root = Server(client)
    site = Site(root)
    port = 8789
    interface = '0.0.0.0'
    if args:
        if ':' in args[0]:
            interface, port = args[0].split(':')
            port = int(port)
        else:
            port = int(args[0])
        local = None
    else:
        try:
            from . import localnode
            local = localnode.Server(port)
        except:
            local = None
    reactor.listenTCP(port, site, interface=interface)
    print('listening on http://%s:%s' % (interface, port))
    if local:
        print('broadcasting location via mdns (you can just use pandora_client client on the other side)')
    client.update_encodes()
    root.queue_available_uploads()
    reactor.run()
    if local:
        del local



import socket
import netifaces
from zeroconf import (
    ServiceBrowser, ServiceInfo, ServiceStateChange, Zeroconf
)

import logging
logger = logging.getLogger(__name__)

service_type = '_pandoraclient._tcp.local.'

def get_broadcast_interfaces():
    return list(set(
        addr['addr']
        for iface in netifaces.interfaces()
        for addr in netifaces.ifaddresses(iface).get(socket.AF_INET, [])
        if addr.get('netmask') != '255.255.255.255' and addr.get('broadcast')
    ))


class Server(object):
    local_info = None
    local_ips = None

    def __init__(self, port):
        self.port = port
        self.name = socket.gethostname().partition('.')[0] + '-%s' % port
        self.local_ips = get_broadcast_interfaces()
        self.zeroconf = {ip: Zeroconf(interfaces=[ip]) for ip in self.local_ips}
        self.register_service()

    def register_service(self):
        if self.local_info:
            for local_ip, local_info in self.local_info:
                self.zeroconf[local_ip].unregister_service(local_info)
            self.local_info = None

        local_name = socket.gethostname().partition('.')[0] + '.local.'
        port = self.port
        desc = {}
        self.local_info = []
        for i, local_ip in enumerate(get_broadcast_interfaces()):
            if i:
                name = '%s-%s.%s' % (self.name, i+1, service_type)
            else:
                name = '%s.%s' % (self.name, service_type)
            local_info = ServiceInfo(service_type, name,
                                     socket.inet_aton(local_ip), port, 0, 0, desc, local_name)
            self.zeroconf[local_ip].register_service(local_info)
            self.local_info.append((local_ip, local_info))

    def __del__(self):
        self.close()

    def close(self):
        if self.local_info:
            for local_ip, local_info in self.local_info:
                try:
                    self.zeroconf[local_ip].unregister_service(local_info)
                except:
                    logger.debug('exception closing zeroconf', exc_info=True)
            self.local_info = None
        if self.zeroconf:
            for local_ip in self.zeroconf:
                try:
                    self.zeroconf[local_ip].close()
                except:
                    logger.debug('exception closing zeroconf', exc_info=True)
            self.zeroconf = None

class LocalNodes(dict):

    def __init__(self):
        self.local_ips = get_broadcast_interfaces()
        self.zeroconf = {ip: Zeroconf(interfaces=[ip]) for ip in self.local_ips}
        self.browse()

    def browse(self):
        self.browser = {
            ip: ServiceBrowser(self.zeroconf[ip], service_type, handlers=[self.on_service_state_change])
            for ip in self.zeroconf
        }

    def __del__(self):
        self.close()

    def close(self):
        if self.zeroconf:
            for local_ip in self.zeroconf:
                try:
                    self.zeroconf[local_ip].close()
                except:
                    logger.debug('exception closing zeroconf', exc_info=True)
            self.zeroconf = None
        for id in list(self):
            self.pop(id, None)

    def on_service_state_change(self, zeroconf, service_type, name, state_change):
        id = name.split('.')[0].split('-')[0]
        if state_change is ServiceStateChange.Added:
            info = zeroconf.get_service_info(service_type, name)
            if info:
                self[id] = 'http://%s:%s' % (socket.inet_ntoa(info.address), info.port)
        elif state_change is ServiceStateChange.Removed:
            logger.debug('remove: %s', id)
            self.pop(id, None)




#!/usr/bin/python
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
# GPL 2012-2016
from __future__ import division, with_statement, print_function, absolute_import

import getpass
from glob import glob
import imp
import json
import math
import os
import socket
import sqlite3
import sys
import tempfile
import time
import pkg_resources

from six.moves.urllib.parse import urlparse
from six.moves import input
from six import string_types

import ox

from . import extract
from . import utils


DEBUG = False

__version__ = pkg_resources.require("pandora_client")[0].version

socket.setdefaulttimeout(300)
CHUNK_SIZE = 1024*1024*5
default_media_cache = os.environ.get('oxMEDIA', os.path.join(utils.basedir(), 'media'))

DOCUMENT_FORMATS = ('jpg', 'pdf', 'png')
sync_extensions = ()

def get_frames(filename, prefix, info, force=False):
    oshash = info['oshash']
    cache = os.path.join(prefix, os.path.join(*utils.hash_prefix(oshash)))
    frames = []
    for pos in utils.video_frame_positions(info['duration']):
        frame_name = '%s.png' % pos
        frame_f = os.path.join(cache, frame_name)
        if force or not os.path.exists(frame_f):
            print(frame_f)
            extract.frame(filename, frame_f, pos)
        frames.append(frame_f)
    return frames

def encode(filename, prefix, profile, info=None, extract_frames=True):
    if not info:
        info = utils.avinfo(filename)
    if 'oshash' not in info:
        return None
    oshash = info['oshash']
    cache = os.path.join(prefix, os.path.join(*utils.hash_prefix(oshash)))
    if info.get('video') and extract_frames:
        frames = get_frames(filename, prefix, info)
    else:
        frames = []
    if info.get('video') or info.get('audio'):
        media_f = os.path.join(cache, profile)
        if not os.path.exists(media_f) or os.stat(media_f).st_size == 0:
            extract.video(filename, media_f, profile, info)
    else:
        print(info)
        print(filename)
        return None
    return {
        'info': info,
        'oshash': oshash,
        'frames': frames,
        'media': media_f
    }

def encode_cmd(filename, prefix, profile, info):
    if not info:
        info = utils.avinfo(filename)
    if 'oshash' not in info:
        return None
    oshash = info['oshash']
    cache = os.path.join(prefix, os.path.join(*utils.hash_prefix(oshash)))
    media_f = os.path.join(cache, profile)
    return extract.video_cmd(filename, media_f, profile, info)

def parse_path(client, path, prefix=None):
    '''
        args:
            path   - path without volume prefix 
            client - Client instance
            prefix - volume prefix
        return:
            return None if file will not be used, dict with parsed item information otherwise
    '''
    if isinstance(path, bytes):
        path = path.decode('utf-8')
    path = path.replace(os.sep, '/')
    parts = path.split('/')
    if len(parts) >= client.folderdepth and parts[client.folderdepth-1] == 'Documents':
        info = ox.movie.parse_path(u'/'.join(
            parts[:client.folderdepth-1] + [parts[-1]]
        ))
    else:
        if len(parts) != client.folderdepth:
            return None
        info = ox.movie.parse_path(path)
    if client.folderdepth == 3:
        info['director'] = []
        info['directorSort'] = []
    return info

def example_path(client):
    return '\t' + (client.folderdepth == 4 and 'L/Last, First/Title (Year)/Title.avi' or 'T/Title/Title.dv')

def ignore_file(client, path):
    filename = os.path.basename(path)
    if filename.startswith('._') \
            or filename in ('.DS_Store', 'Thumbs.db') \
            or filename.endswith('~') \
            or 'Extras' + os.sep in path \
            or 'Versions' + os.sep in path \
            or not os.path.exists(path) \
            or os.stat(path).st_size == 0:
        return True
    return False

def is_oshash(oshash):
    try:
        int(oshash, 16)
    except:
        return False
    return len(oshash) == 16

def hide_cursor():
    sys.stdout.write("\033[?25l")
    sys.stdout.flush()

def show_cursor():
    sys.stdout.write("\033[?25h")
    sys.stdout.flush()

class Client(object):
    _configfile = None

    def __init__(self, config, offline=False):
        if isinstance(config, string_types):
            self._configfile = os.path.expanduser(config)
            with open(config) as f:
                try:
                    self._config = json.load(f)
                except ValueError:
                    print("Failed to parse config at", config)
                    sys.exit(1)
            base = self._config.get('plugin.d', os.path.join(utils.basedir(), 'client.d'))
            self.load_plugins(base)
        else:
            self._config = config
        if not self._config['url'].endswith('/'):
            self._config['url'] = self._config['url'] + '/'
        self.resolutions = list(reversed(sorted(self._config.get('resolutions', [480]))))
        self.format = self._config.get('format', 'webm')
        self.importFrames = False

        if not offline:
            self.online()

        conn, c = self._conn()

        c.execute(u'CREATE TABLE IF NOT EXISTS setting (key varchar(1024) unique, value text)')

        if int(self.get('version', 0)) < 1:
            self.set('version', 1)
            db = [
                u'''CREATE TABLE IF NOT EXISTS file (
                                path varchar(1024) unique,
                                oshash varchar(16),
                                atime FLOAT,
                                ctime FLOAT,
                                mtime FLOAT,
                                size INT,
                                info TEXT,
                                created INT,
                                modified INT,
                                deleted INT)''',
                u'CREATE INDEX IF NOT EXISTS path_idx ON file (path)',
                u'CREATE INDEX IF NOT EXISTS oshash_idx ON file (oshash)',
            ]
            for i in db:
                c.execute(i)
            conn.commit()
        if int(self.get('version', 0)) < 2:
            self.set('version', 2)
            db = [
                u'''CREATE TABLE IF NOT EXISTS encode (
                                oshash varchar(16),
                                site varchar(255))''',
                u'CREATE INDEX IF NOT EXISTS upload_site_idx ON encode (site)',
            ]
            for i in db:
                c.execute(i)
            conn.commit()
        if int(self.get('version', 0)) < 3:
            self.set('version', 3)
            db = [
                u'ALTER TABLE file add sha1 varchar(42)'
            ]
            for i in db:
                c.execute(i)
            conn.commit()
        if int(self.get('version', 0)) < 4:
            self.set('version', 4)
            db = [
                u'ALTER TABLE encode add status varchar(255)',
                u'CREATE INDEX IF NOT EXISTS encode_status_idx ON encode (status)',
                u'ALTER TABLE encode ADD modified INT DEFAULT 0',
            ]
            for i in db:
                c.execute(i)
            conn.commit()
        conn.close()

    def load_plugins(self, base=os.path.join(utils.basedir(), 'client.d')):
        global parse_path, example_path, ignore_file, sync_extensions, encode
        base = os.path.expanduser(base)
        for path in sorted(glob('%s%s*.py' % (base, os.sep))):
            with open(path) as fp:
                module = imp.load_source(os.path.basename(path).split('.')[0], base, fp)
                if hasattr(module, 'parse_path'):
                    parse_path = module.parse_path
                if hasattr(module, 'example_path'):
                    example_path = module.example_path
                if hasattr(module, 'ignore_file'):
                    ignore_file = module.ignore_file
                if hasattr(module, 'sync_extensions'):
                    sync_extensions = module.sync_extensions
                if hasattr(module, 'encode'):
                    encode = module.encode

    def _conn(self):
        db_conn = self._config['cache']
        if isinstance(db_conn, bytes):
            db_conn = db_conn.decode('utf-8')
        db_conn = os.path.expanduser(db_conn)
        if not os.path.exists(os.path.dirname(db_conn)):
            os.makedirs(os.path.dirname(db_conn))
        conn = sqlite3.connect(db_conn, timeout=10)
        return conn, conn.cursor()

    def media_cache(self):
        path = self._config.get('media-cache', default_media_cache)
        if path is None:
            path = '/tmp/pandora_client_media_cache'
        return os.path.expanduser(path)

    def get(self, key, default=None):
        conn, c = self._conn()
        c.execute(u'SELECT value FROM setting WHERE key = ?', (key, ))
        value = default
        for row in c:
            value = row[0]
            break
        conn.close()
        return value

    def set(self, key, value):
        conn, c = self._conn()
        c.execute(u'INSERT OR REPLACE INTO setting values (?, ?)', (key, str(value)))
        conn.commit()
        conn.close()

    def info(self, oshash):
        conn, c = self._conn()
        c.execute(u'SELECT info FROM file WHERE oshash = ?', (oshash, ))
        info = None
        for row in c:
            info = json.loads(row[0])
            break
        conn.close()
        return info

    def get_info(self, oshash, prefix=None):
        if prefix:
            prefixes = [prefix]
        else:
            prefixes = self.active_volumes().values()
        prefixes = [p.decode('utf-8') if isinstance(prefix, bytes) else p for p in prefixes]
        _info = self.info(oshash)
        for path in self.path(oshash):
            for prefix in prefixes:
                if path.startswith(prefix) and os.path.exists(path):
                    path = path[len(prefix):]
                    i = parse_path(self, path, prefix)
                    if i:
                        _info.update(i)
                        return _info
                    else:
                        print('failed to parse', path)
                        return

    def get_info_for_ids(self, ids, prefix=None):
        info = {}
        for oshash in ids:
            i = self.get_info(oshash, prefix)
            if i:
                info[oshash] = i
        return info

    def path(self, oshash):
        conn, c = self._conn()
        c.execute(u'SELECT path FROM file WHERE oshash = ?', (oshash, ))
        paths = set()
        for row in c:
            path = row[0]
            paths.add(path)
        conn.close()
        return list(paths)

    def online(self):
        self.api = API(self._config['url'], media_cache=self.media_cache())
        self.api.DEBUG = DEBUG
        if self.signin():
            self.resolutions = list(reversed(sorted(self.api.site['video']['resolutions'])))
            self.format = self.api.site['video']['formats'][0]
            self.importFrames = self.api.site['media'].get('importFrames')
        self.folderdepth = self._config.get('folderdepth', self.api.site['site'].get('folderdepth', 3))

    def signin(self):
        if 'username' in self._config:
            r = self.api.signin(username=self._config['username'], password=self._config['password'])
            if r['status']['code'] == 200 and 'errors' not in r['data']:
                self.user = r['data']['user']
            else:
                self.user = False
                if DEBUG:
                    print(r)
                print('\nlogin failed! check config\n\n')
                sys.exit(1)
            r = self.api.init()
            if r['status']['code'] == 200:
                self.api.site = r['data']['site']
            else:
                print("\n init failed.", r['status'])
                sys.exit(1)
            return True

    def set_encodes(self, site, files):
        conn, c = self._conn()
        c.execute(u'DELETE FROM encode WHERE site = ?', (site, ))
        conn.commit()
        conn.close()
        self.add_encodes(site, files)

    def get_encodes(self, site, status=''):
        conn, c = self._conn()
        sql = u'SELECT oshash FROM encode WHERE site = ? AND status = ?'
        args = [site, status]
        c.execute(sql, tuple(args))
        encodes = [row[0] for row in c]
        conn.close()
        return encodes

    def add_encodes(self, site, files):
        conn, c = self._conn()
        for oshash in files:
            c.execute(u'INSERT INTO encode VALUES (?, ?, ?, 0)', (oshash, site, ''))
        conn.commit()
        conn.close()

    def update_encodes(self, add=False):
        # send empty list to get updated list of requested info/files/data
        site = self._config['url']
        post = {'info': {}}
        r = self.api.update(post)
        files = r['data']['data']
        if add:
            conn, c = self._conn()
            sql = u'SELECT oshash FROM encode WHERE site = ?'
            c.execute(sql, (site, ))
            known = [row[0] for row in c]
            conn.close()
            files = list(set(files) - set(known))
            if files:
                self.add_encodes(site, files)
        else:
            self.set_encodes(site, files)

    def scan_file(self, path, rescan=False):
        conn, c = self._conn()

        update = True
        modified = time.mktime(time.localtime())
        created = modified

        if isinstance(path, bytes):
            path = path.decode('utf-8')

        sql = u'SELECT atime, ctime, mtime, size, created, info FROM file WHERE deleted < 0 AND path=?'
        c.execute(sql, [path])
        stat = os.stat(path)
        for row in c:
            if stat.st_atime == row[0] and stat.st_ctime == row[1] and stat.st_mtime == row[2] and stat.st_size == row[3]:
                created = row[4]
                info = json.loads(row[5])
                update = False
            break
        if update or rescan:
            info = utils.avinfo(path, cached=not rescan)
            if info['size'] > 0:
                oshash = info['oshash']
                sha1 = None
                deleted = -1
                t = (path, oshash, stat.st_atime, stat.st_ctime, stat.st_mtime,
                     stat.st_size, json.dumps(info), created, modified, deleted, sha1)
                c.execute(u'INSERT OR REPLACE INTO file values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)',
                          t)
                conn.commit()
        conn.close()
        return 'error' not in info

    def get_resolution(self, info):
        height = info['video'][0]['height'] if info.get('video') else None
        for resolution in sorted(self.resolutions):
            if height and height <= resolution:
                return resolution
        return resolution

    def profile(self, info):
        resolution = self.get_resolution(info)
        profile = '%sp.%s' % (resolution, self.format)
        return profile

    def cmd(self, args):
        filename = args[0]
        if len(filename) == 16:
            path = self.path(filename)
        else:
            path = [filename]
        for p in path:
            if os.path.exists(p):
                info = utils.avinfo(p)
                profile = self.profile(info)
                cmds = encode_cmd(p, self.media_cache(), profile, info)
                output = []
                for cmd in cmds:
                    cmd = [' ' in c and u'"%s"' % c or c for c in cmd]
                    output.append(u' '.join(cmd))
                print(' && '.join(output))

    def save_config(self):
        if not self._configfile:
            raise Exception('Can not save temporary config')
        with open(self._configfile, 'w') as f:
            json.dump(self._config, f, indent=2)

    def config(self, args):
        print("Current Config:\n  User: %s\n  URL: %s\n" % (self._config['username'], self._config['url']))
        print("Leave empty to keep current value\n")
        username = input('Username: ')
        if username:
            self._config['username'] = username
        password = getpass.getpass('Password: ')
        if password:
            self._config['password'] = password
        url = input('Pan.do/ra URL (i.e. https://pandora.local/api/): ')
        if url:
            self._config['url'] = url
        self.save_config()
        print("\nconfiguration updated.")
        self.install_programs()

    def install_programs(self, args=[]):
        update = 'update' in args
        # install required programs
        if sys.platform == 'darwin':
            osext = 'macosx'
        elif sys.platform.startswith('win'):
            osext = 'exe'
        else:
            osext = 'linux'
        bindir = os.path.join(utils.basedir(), 'bin')
        ox.makedirs(bindir)
        for p in ('ffmpeg', 'ffmpeg2theora'):
            path = os.path.join(bindir, p)
            if sys.platform.startswith('win'):
                p += '.exe'
            if not os.path.exists(path) or update:
                print("installing %s in %s" % (p, bindir))
                ox.net.save_url('http://firefogg.org/bin/%s.%s' % (p, osext), path, True)
                os.chmod(path, 0o755)

    def add_volume(self, args):
        usage = "Usage: %s add_volume name path" % sys.argv[0]
        if len(args) != 2:
            print(usage)
            sys.exit(1)
        name = args[0]
        path = os.path.abspath(args[1])
        if not path.endswith(os.sep):
            path += os.sep
        if os.path.isdir(path):
            if name in self._config['volumes']:
                print("updated %s to %s" % (name, path))
            else:
                print("added %s %s" % (name, path))
            self._config['volumes'][name] = path
            self.save_config()
        else:
            print("'%s' does not exist" % path)
            print(usage)
            sys.exit(1)

    def active_volumes(self):
        volumes = {}
        for name in sorted(self._config['volumes']):
            path = self._config['volumes'][name]
            path = os.path.normpath(path)
            if not path.endswith(os.sep):
                path += os.sep
            if isinstance(path, bytes):
                path = path.decode('utf-8')
            if os.path.exists(path):
                volumes[name] = path
        return volumes

    def scan(self, args):
        rescan = 'rescan' in args
        print("checking for new files ...")
        volumes = self.active_volumes()
        for name in sorted(volumes):
            path = volumes[name]
            conn, c = self._conn()
            c.execute(u'SELECT path FROM file WHERE path LIKE ? AND deleted < 0', [u"%s%%" % path])
            known_files = [r[0] for r in c.fetchall()]
            conn.close()

            files = []
            unknown = []
            ignored = []
            unsupported = []
            for dirpath, dirnames, filenames in os.walk(path, followlinks=True):
                if filenames:
                    for filename in sorted(filenames):
                        file_path = os.path.join(dirpath, filename)
                        if not ignore_file(self, file_path):
                            files.append(file_path)
                        else:
                            ignored.append(file_path)
            for f in files:
                if not parse_path(self, f[len(path):], path):
                    unknown.append(f)

            files = sorted(set(files) - set(unknown))

            for f in files:
                if os.path.splitext(f)[-1] not in sync_extensions and not self.scan_file(f, rescan):
                    unsupported.append(f)

            if unknown:
                example = example_path(self)
                print('Files need to be in a folder structure like this:\n%s\n' % example)
                print('The following files do not fit into the folder structure and will not be synced:')
                print('\t' + '\n\t'.join([f[len(path):] for f in sorted(unknown)]))
                print('')

            if unsupported:
                files = sorted(set(files) - set(unsupported))
                print('The following files are in an unsupported format and will not be synced:')
                print('\t' + '\n\t'.join([f[len(path):] for f in sorted(unsupported)]))
                print('')
            '''
            '''

            deleted_files = list(filter(lambda f: f not in files, known_files))
            new_files = list(filter(lambda f: f not in known_files, files))
            conn, c = self._conn()
            if deleted_files:
                deleted = time.mktime(time.localtime())
                for f in deleted_files:
                    c.execute(u'UPDATE file SET deleted=? WHERE path=?', (deleted, f))
                conn.commit()
            conn.close()

            '''
            print("scanned volume %s: %s files, %s new, %s deleted, %s ignored, %s unsupported" % (
                  name, len(files), len(new_files), len(deleted_files), len(ignored), len(unsupported)))
            '''
            print("scanned volume %s: %s files, %s new, %s deleted, %s ignored" % (
                  name, len(files), len(new_files), len(deleted_files), len(ignored)))

    def extract(self, args):
        if args:
            if args[0] == 'offline':
                files = self.get_encodes(self._config['url'])
            elif args[0] == 'all':
                files = []
                for name in self._config['volumes']:
                    path = self._config['volumes'][name]
                    path = os.path.normpath(path)
                    if not path.endswith(os.sep):
                        path += os.sep
                    if os.path.exists(path):
                        files += self.files(path)['info']
            else:
                files = [f if len(f) == 16 else ox.oshash(f) for f in args]
        else:
            if not self.user:
                print("you need to login or run pandora_client extract offline")
                return
            self.update_encodes()
            files = self.get_encodes(self._config['url'])

        for oshash in files:
            info = self.info(oshash)
            if 'error' not in info:
                for path in self.path(oshash):
                    if os.path.exists(path):
                        profile = self.profile(info)
                        i = encode(path, self.media_cache(), profile, info,
                                   self.importFrames)
                        break

    def sync(self, args):
        if not self.user:
            print("you need to login")
            return
        conn, c = self._conn()

        volumes = self.active_volumes()
        if not volumes:
            print("no active volumes found")
            return

        for name in sorted(volumes):
            prefix = volumes[name]
            files = self.files(prefix)
            post = {}
            post['files'] = files['files']
            post['volume'] = name
            print('sending list of files in %s (%s total)' % (name, len(post['files'])))
            r = self.api.later('update', post)
            # send empty list to get updated list of requested info/files/data
            post = {'info': {}}
            r = self.api.update(post)

            if r['data']['info']:
                r = self.update_info(r['data']['info'], prefix)

        if 'data' not in r:
            print(r)
            return

        if r['data']['data']:
            files = []
            for f in r['data']['data']:
                for path in self.path(f):
                    if os.path.exists(path):
                        files.append(path)
                        break
            if files:
                print('\ncould encode and upload %s videos:\n' % len(files))
                print('\n'.join(files))
        if r['data']['file']:
            files = []
            for f in r['data']['file']:
                for path in self.path(f):
                    if os.path.exists(path):
                        files.append(path)
                        break
            if files:
                print('\ncould upload %s subtitles:\n' % len(files))
                print('\n'.join(files))

    def upload(self, args):
        if not self.user:
            print("you need to login")
            return

        documents = []
        if args:
            data = []
            for arg in args:
                if os.path.exists(arg):
                    oshash = ox.oshash(arg)
                    self.scan_file(arg)
                    r = self.api.findMedia({'query': {
                        'conditions': [{'key': 'oshash', 'value': oshash}]
                    }})['data']['items']
                    if r == 0:
                        info = self.info(oshash)
                        filename = os.path.basename(arg)
                        r = self.api.addMedia({
                            'id': oshash,
                            'info': info,
                            'filename': filename
                        })
                    data.append(oshash)
                elif not is_oshash(arg):
                    print('file not found "%s"' % arg)
                    sys.exit(1)
                else:
                    data.append(arg)
            files = []
            info = []
        else:
            if not self.active_volumes():
                print("no volumes found, mount volumes and run again")
                return
            # send empty list to get updated list of requested info/files/data
            post = {'info': {}}
            r = self.api.update(post)
            data = r['data']['data']
            files = r['data']['file']
            info = r['data']['info']
            documents = self._get_documents()

        if info:
            r = self.update_info(info)
            data = r['data']['data']
            files = r['data']['file']

        if files:
            print('uploading %s files' % len(files))
            for oshash in files:
                for path in self.path(oshash):
                    if os.path.exists(path):
                        self.api.uploadData(path, oshash)
                        break
        if documents:
            _documents = []
            for oshash, item in documents:
                for path in self.path(oshash):
                    if os.path.exists(path):
                        _documents.append([path, item])
                        break
            print('uploading %s documents' % len(_documents))
            for path, item in _documents:
                self._add_document(path, item)

        if data:
            print('encoding and uploading %s videos' % len(data))
            for oshash in data:
                data = {}
                info = self.info(oshash)
                if info and 'error' not in info:
                    for path in self.path(oshash):
                        if os.path.exists(path):
                            if not self.api.uploadVideo(path, data,
                                                        self.profile(info), info):
                                print('video upload failed, giving up, please try again')
                                return
                            if 'rightsLevel' in self._config:
                                r = self.api.find({'query': {
                                    'conditions': [
                                        {'key': 'oshash', 'value': oshash, 'operator': '=='}
                                    ],
                                    'keys': ['id'],
                                    'range': [0, 1]
                                }})
                                if r['data']['items']:
                                    item = r['data']['items'][0]['id']
                                    r = self.api.edit({
                                        'item': item,
                                        'rightsLevel': self._config['rightsLevel']
                                    })
                            break

    def update_info(self, info, prefix=None):
        if info:
            print('sending info for %d files' % len(info))
            post = {'info': {}, 'upload': True}
            post['info'] = self.get_info_for_ids(info, prefix)
            r = self.api.later('update', post)
            # send empty list to get updated list of requested info/files/data
            post = {'info': {}}
            r = self.api.update(post)
        return r

    def upload_frames(self, args):
        if not self.user:
            print("you need to login")
            return
        for oshash in args:
            info = self.info(oshash)
            if info and 'error' not in info:
                for path in self.path(oshash):
                    if os.path.exists(path):
                        frames = get_frames(path, self.api.media_cache, info, True)
                        i = {
                            'info': info,
                            'oshash': oshash,
                            'frames': frames,
                        }
                        r = self.api.uploadFrames(i, {})
                        if r.get('status', {}).get('code') != 200:
                            print(r)

    def _get_documents(self):
        files = []
        for volume in self.active_volumes():
            query = {
                'conditions': [
                    {'key': 'list', 'value': volume, 'operator': '=='},
                    {
                        'conditions': [
                            {'key': 'filename', 'operator': '', 'value': value}
                            for value in DOCUMENT_FORMATS
                        ],
                        'operator': '|'
                    }
                ],
                'operator': '&'
            }
            n = self.api.findMedia({'query': query})['data']['items']
            if n:
                o = 0
                chunk = 5000
                while o < n:
                    files += [f for f in self.api.findMedia({
                        'query': query,
                        'keys': ['item', 'id', 'extension'],
                        'range': [o, o+chunk]
                    })['data']['items'] if f['extension'] in DOCUMENT_FORMATS]
                    o += chunk
        missing = list(set((f['id'], f['item']) for f in files))
        available = set()
        total = len(missing)
        ids = [m[0] for m in missing]
        o = 0
        chunk = 1000
        while o < len(ids):
            for d in self.api.findDocuments({
                'query': {
                    'conditions': [
                        {'key': 'oshash', 'operator': '==', 'value': id}
                        for id in ids[o:o+chunk]
                    ],
                    'operator': '|'
                },
                'keys': ['oshash'],
                'range': [0, chunk]
            })['data']['items']:
                available.add(d['oshash'])
            o += chunk
        missing = [m for m in missing if m[0] not in available]
        return missing

    def find_document(self, oshash):
        r = self.api.findDocuments({
            'keys': ['id'],
            'query': {
                'conditions': [
                    {'key': 'oshash', 'value': oshash, 'operator': '=='}
                ],
                'operator': '&'
            }
        })
        if r['data']['items']:
            return r['data']['items'][0]['id']
        return None

    def _add_document(self, f, item=None):
        if f.split('.')[-1].lower() not in DOCUMENT_FORMATS:
            print('skip, not a document', f)
            return False
        oshash = ox.oshash(f)
        did = self.find_document(oshash)
        if not did:
            url = '%supload/document/' % self._config['url']
            did = self.api.upload_chunks(url, f, {
                'filename': os.path.basename(f)
            })
        if did and item:
            r = self.api.addDocument({
                'id': did,
                'item': item
            })
        return did

    def upload_document(self, args):
        if not self.user:
            print("you need to login")
            return
        for f in args:
            r = self._add_document(f)
            if not r:
                print('unsupported format', f)
                continue

    def files(self, prefix):
        if not prefix.endswith('/'):
            prefix += '/'
        conn, c = self._conn()
        files = {}
        files['info'] = {}
        files['files'] = []
        sql = u'SELECT path, oshash, info, atime, ctime, mtime FROM file WHERE deleted < 0 AND path LIKE ? ORDER BY path'
        t = [u"%s%%" % prefix]
        c.execute(sql, t)
        for row in c:
            path = row[0]
            oshash = row[1]
            info = json.loads(row[2])
            if 'error' not in info:
                for key in ('atime', 'ctime', 'mtime', 'path'):
                    if key in info:
                        del info[key]
                files['info'][oshash] = info
                files['files'].append({
                    'oshash': oshash,
                    'path': path[len(prefix):],
                    'atime': row[3],
                    'ctime': row[4],
                    'mtime': row[5],
                })
        conn.close()
        return files

    def clean(self, args):
        if os.path.exists(self.api.media_cache):
            if args and args[0] == 'all':
                print("remove all cached videos in", self.api.media_cache)

                # if os.path.exists(self.api.media_cache):
                #    shutil.rmtree(self.api.media_cache)
            else:
                nothing = False
                for root, folders, files in os.walk(self.api.media_cache):
                    for f in sorted(files):
                        f = os.path.join(root, f)
                        if f.endswith('.webm'):
                            oshash = os.path.dirname(f)[len(self.api.media_cache):].replace('/', '')
                            remove = True
                            for path in self.path(oshash):
                                if os.path.exists(path):
                                    remove = False
                                    break
                            if remove:
                                nothing = False
                                os.unlink(f)
                if nothing and folders:
                    print("No unused files found in cache, run \"%s clean all\" to remove the entire cache" % sys.argv[0])
                else:
                    utils.cleanup_tree(self.api.media_cache)

    def import_srt(self, args):
        '''
            import srt as annotations, usage:
                pandora_client ITEMID layername /path/to/transcript.srt
            i.e. 
                pandora_client ABC transcripts /path/to/transcript.srt
        '''
        if not args:
            print('Usage: pandora_client ABC transcripts /path/to/transcript.srt')
            sys.exit(1)
        item = args[0]
        layer = args[1]
        filename = args[2]
        layers = [l['id'] for l in self.api.site['layers']]
        if layer not in layers:
            print("invalid layer name, choices are: ", ', '.join(layers))
            sys.exit(1)
        if filename.endswith('.vtt'):
            load = ox.vtt.load
        else:
            load = ox.srt.load
        annotations = [{
            'in': s['in'],
            'out': s['out'],
            'value': s['value'].replace('\n', '<br>\n') if layer == 'subtitles' else s['value'],
        } for s in load(filename)]
        r = self.api.addAnnotations({
            'item': item,
            'layer': layer,
            'annotations': annotations
        })
        if r['status']['code'] == 400:
            print('failed')
            sys.exit(1)
        if r['status']['code'] == 403:
            print('permission deinied')
            sys.exit(1)
        elif r['status']['code'] == 404:
            print('item not found')
            sys.exit(1)

    def server(self, args):
        from . import server
        server.run(self, args)

    def client(self, args):
        threads = [t.split('=')[-1] for t in args if t.startswith('c=')]
        if threads:
            threads = int(threads[0])
        else:
            threads = 1
        args = [a for a in args if not a.startswith('c=')]
        urls = [u for u in args if u.startswith('http:')]
        name = [u for u in args if u not in urls]
        if not name:
            name = '%s-%s' % (socket.gethostname(), int(time.time()))
        else:
            name = name[0]
        if not urls:
            from . import localnode
            nodes = localnode.LocalNodes()
            time.sleep(1)
            found = len(nodes)
            if not found:
                print('usage: %s client <server_url>\n\ti.e. %s client http://192.168.1.1:8789' % (sys.argv[0], sys.argv[0]))
                sys.exit(1)
            elif found > 1:
                print('found multiple servers, please select one, your options are:')
                for id, url in nodes.items():
                    print('\t%s client %s' % (sys.argv[0], url))
                sys.exit(1)

            else:
                for id, url in nodes.items():
                    break
                print('connecting to %s (%s)' % (id, url))
        else:
            url = urls[0]
        from . import client
        c = client.DistributedClient(url, name, threads)
        c.run()

class API(ox.API):
    __name__ = 'pandora_client'
    __version__ = __version__

    def __init__(self, url, cj=None, media_cache=None):
        super(API, self).__init__(url, cj)

        self.media_cache = media_cache
        if not self.media_cache:
            self.media_cache = os.path.expanduser(default_media_cache)
        netloc = urlparse(self.url).netloc
        tmp = tempfile.gettempdir()
        self._resume_file = os.path.join(tmp, 'pandora_client.%s.%s.json' % (os.environ.get('USER'), netloc))

        if hasattr(self, 'taskStatus') and not hasattr(self, 'getTaskStatus'):
            self.getTaskStatus = self.taskStatus

    def later(self, action, data, interval=5):
        t = r = getattr(self, action)(data)
        if r['status']['code'] == 200:
            # wait for async task to finish
            if 'taskId' in r['data']:
                t = self.getTaskStatus(task_id=r['data']['taskId'])
                print('waiting for server ...')
                while t['data'].get('status') == 'PENDING':
                    time.sleep(interval)
                    t = self.getTaskStatus(task_id=r['data']['taskId'])
        return t

    def uploadFrames(self, i, data):
        # upload frames
        if self.site['media'].get('importFrames') and i['frames']:
            form = ox.MultiPartForm()
            form.add_field('action', 'upload')
            form.add_field('id', i['oshash'])
            for key in data:
                form.add_field(key, data[key])
            for frame in i['frames']:
                fname = os.path.basename(frame)
                if os.path.exists(frame):
                    form.add_file('frame', fname, open(frame, 'rb'))
            r = self._json_request(self.url, form)
            return r

    def uploadVideo(self, filename, data, profile, info=None):
        i = encode(filename, self.media_cache, profile, info,
                   self.site['media'].get('importFrames'))
        if not i:
            print("failed")
            return

        # upload frames
        r = self.uploadFrames(i, data)

        # upload media
        if os.path.exists(i['media']):
            size = ox.format_bytes(os.path.getsize(i['media']))
            name = os.path.basename(filename)
            print(u"uploading %s of %s (%s)" % (profile, name, size))
            url = self.url + 'upload/?profile=%s&id=%s' % (profile, i['oshash'])
            if not self.upload_chunks(url, i['media'], data):
                if DEBUG:
                    print("failed")
                return False
        else:
            print("Failed")
            return False
        return True

    def uploadData(self, filename, oshash):
        if DEBUG:
            print('upload', filename)
        form = ox.MultiPartForm()
        form.add_field('action', 'upload')
        form.add_field('id', str(oshash))
        fname = os.path.basename(filename)
        if not isinstance(fname, bytes):
            fname = fname.encode('utf-8')
        form.add_file('file', fname, open(filename, 'rb'))
        r = self._json_request(self.url, form)
        return r

    def upload_chunks(self, url, filename, data=None):
        form = ox.MultiPartForm()
        resume = None
        if self._resume_file and os.path.exists(self._resume_file):
            with open(self._resume_file) as f:
                try:
                    resume = json.load(f)
                except:
                    resume = {}
            if resume.get('chunkUploadUrl') != url:
                resume = None
        if resume:
            data = resume
        else:
            for key in data:
                form.add_field(key, data[key])
            data = self._json_request(url, form)

        print(filename)
        hide_cursor()

        def full_url(path):
            if path.startswith('/'):
                u = urlparse(url)
                path = '%s://%s%s' % (u.scheme, u.netloc, path)
            return path

        result_url = full_url(data.get('url'))
        if 'uploadUrl' in data:
            uploadUrl = full_url(data['uploadUrl'])
            f = open(filename, 'rb')
            fsize = os.stat(filename).st_size
            done = 0
            start = time.mktime(time.localtime())
            if 'offset' in data and data['offset'] < fsize:
                done = data['offset']
                f.seek(done)
                resume_offset = done
            else:
                resume_offset = 0
            chunk = f.read(CHUNK_SIZE)
            fname = os.path.basename(filename)
            if not isinstance(fname, bytes):
                fname = fname.encode('utf-8')
            while chunk:
                elapsed = time.mktime(time.localtime()) - start
                remaining = ""
                if done:
                    r = math.ceil((elapsed / (done/(fsize-resume_offset)) - elapsed)/60) * 60 * 1000 
                    r = ox.format_duration(r, milliseconds=False, verbosity=2)
                    if r:
                        remaining = ", %s remaining" % r
                msg = '%0.2f%% %s of %s done%s' % (
                    100 * done/fsize, ox.format_bytes(done), ox.format_bytes(fsize), remaining)
                print(''.join([msg, ' ' * (80-len(msg)), '\r']), end='')
                sys.stdout.flush()
                form = ox.MultiPartForm()
                form.add_file('chunk', fname, chunk)
                if len(chunk) < CHUNK_SIZE or f.tell() == fsize:
                    form.add_field('done', '1')
                form.add_field('offset', str(done))
                try:
                    data = self._json_request(uploadUrl, form)
                except KeyboardInterrupt:
                    print("\ninterrupted by user.")
                    sys.exit(1)
                except:
                    print("uploading chunk failed, will try again in 5 seconds\r", end='')
                    sys.stdout.flush()
                    if DEBUG:
                        print('\n', uploadUrl)
                        import traceback
                        traceback.print_exc()
                    data = {'result': -1}
                    time.sleep(5)
                if data and 'status' in data:
                    if data['status']['code'] == 403:
                        print("login required")
                        return False
                    if data['status']['code'] != 200:
                        print("request returned error, will try again in 5 seconds")
                        if DEBUG:
                            print(data)
                        time.sleep(5)
                if data and data.get('result') == 1:
                    done += len(chunk)
                    if data.get('offset') not in (None, done):
                        print('server offset out of sync, continue from', data['offset'])
                        done = data['offset']
                        f.seek(done)
                    if self._resume_file:
                        with open(self._resume_file, 'w') as r:
                            json.dump({
                                'uploadUrl': uploadUrl,
                                'chunkUploadUrl': url,
                                'url': result_url,
                                'offset': done
                            }, r, indent=2)
                    chunk = f.read(CHUNK_SIZE)
            if self._resume_file and os.path.exists(self._resume_file):
                os.unlink(self._resume_file)
                resume = None
            if result_url:
                print(result_url + (' ' * (80-len(result_url))))
            else:
                print(' ' * 80)
            print('')
            show_cursor()
            if data and 'result' in data and data.get('result') == 1:
                return data.get('id', True)
            else:
                return False
        else:
            if DEBUG:
                if 'status' in data and data['status']['code'] == 401:
                    print("login required")
                else:
                    print("failed to upload file to", url)
                    print(data)
        return False




# encoding: utf-8
# vi:si:et:sw=4:sts=4:ts=4
from __future__ import division, with_statement, print_function, absolute_import

import json
import os
import socket
import subprocess
import sys
import threading
import time

import requests

from . import extract

class Encoding:

    def worker(self):
        while True:
            if not self.client.next():
                return

    def __init__(self, client, threads=1):
        self.client = client
        self.threads = []
        for i in range(threads):
            t = threading.Thread(target=self.worker)
            t.start()
            self.threads.append(t)

    def join(self):
        for t in self.threads:
            t.join()


class DistributedClient:

    interrupted = False

    def __init__(self, url, name, threads):
        self.url = url
        self.name = name
        self.threads = threads
        self.supported_formats = extract.supported_formats()

    def ping(self, oshash):
        try:
            url = '%s/ping/%s/%s' % (self.url, oshash, self.name)
            requests.get(url)
        except:
            print('cound not ping server')

    def status(self, oshash, status):
        url = '%s/status/%s' % (self.url, oshash)
        requests.post(url, {'error': status})

    def upload(self, oshash, path):
        url = '%s/upload/%s' % (self.url, oshash)
        with open(path, 'rb') as f:
            requests.put(url, f)

    def next(self):
        url = '%s/next' % self.url
        data = requests.get(url).json()
        if 'oshash' in data:
            return self.encode(data['oshash'], data['cmd'], data['output'])
        return False

    def encode(self, oshash, cmds, output):
        if not isinstance(cmds[0], list):
            cmds = [cmds]
        for cmd in cmds:
            cmd[0] = extract.command('ffmpeg')
            if self.threads > 1:
                cmd = cmd[:1] + ['-nostats', '-loglevel', 'error'] + cmd[1:]
            if 'webm' in cmd and not self.supported_formats['webm']:
                print("ffmpeg is compiled without WebM support")
                return
            elif cmd[-1].endswith('.mp4') and not self.supported_formats['webm']:
                print("ffmpeg is compiled without H.264 support")
                return
            try:
                if self.threads > 1:
                    print('encode', oshash)
                p = subprocess.Popen(cmd)
                r = None
                n = 0
                while True:
                    r = p.poll()
                    if r is None:
                        if n % 60 == 0:
                            self.ping(oshash)
                            n = 0
                        time.sleep(2)
                        n += 2
                    else:
                        break
            except KeyboardInterrupt:
                p.kill()
                #encoding was stopped, put back in queue
                self.status(oshash, '')
                if os.path.exists(output):
                    os.unlink(output)
                if self.threads > 1:
                    self.interrupted = True
                    return False
                sys.exit(1)
            if r != 0:
                break
        if r == 0:
            if self.threads > 1:
                print('ok', oshash)
            self.upload(oshash, output)
        else:
            if self.threads > 1:
                print('error', oshash)
            self.status(oshash, 'failed')
        if os.path.exists(output):
            os.unlink(output)
        return True

    def run(self):
        if self.threads > 1:
            enc = Encoding(self, self.threads)
            enc.join()
        else:
            self.run_single()

    def run_single(self):
        new = True
        while True:
            if not self.next():
                if new:
                    new = False
                    print("currently no more files to encode, ctrl-c to quit")
                try:
                    time.sleep(60)
                except KeyboardInterrupt:
                    return
            else:
                new = True


if __name__ == '__main__':
    url = 'http://127.0.0.1:8789'
    if len(sys.args) == 0:
        name = socket.gethostname()
    else:
        name = sys.argv[1]
    c = DistributedClient(url, name)
    c.run()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
# GPL 2017
from __future__ import division, print_function, absolute_import

from glob import glob
import os
import shutil
import subprocess
import sys

import ox

from .utils import AspectRatio, run_command, basedir

FFMPEG_SUPPORTS_VP9 = False

def command(program):
    local = os.path.join(basedir(), 'bin', program)
    if sys.platform.startswith('win'):
        local += '.exe'
    if os.path.exists(local):
        program = local
    return program

def frame(video, target, position):
    fdir = os.path.dirname(target)
    if fdir and not os.path.exists(fdir):
        os.makedirs(fdir)

    # ffmpeg
    pre = position - 2
    if pre < 0:
        pre = 0
    else:
        position = 2
    cmd = [command('ffmpeg'), '-y', '-ss', str(pre), '-i', video, '-ss', str(position),
           '-vf', 'scale=iw*sar:ih',
           '-an', '-vframes', '1', target]
    r = run_command(cmd)
    return r == 0

def supported_formats():
    ffmpeg = command('ffmpeg')
    if ffmpeg is None:
        return None
    p = subprocess.Popen([ffmpeg, '-codecs'],
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    stdout, stderr = p.communicate()
    stdout = stdout.decode('utf-8')
    return {
        'ogg': 'libtheora' in stdout and 'libvorbis' in stdout,
        'webm': 'libvpx' in stdout and 'libvorbis' in stdout,
        'vp8': 'libvpx' in stdout and 'libvorbis' in stdout,
        'vp9': 'libvpx-vp9' in stdout and 'libopus' in stdout,
        'mp4': 'libx264' in stdout and 'DEA.L. aac' in stdout,
        'h264': 'libx264' in stdout and 'DEA.L. aac' in stdout,
    }

def video_cmd(video, target, profile, info):

    if not os.path.exists(target):
        ox.makedirs(os.path.dirname(target))

    '''
        look into
            lag
            mb_static_threshold
            qmax/qmin
            rc_buf_aggressivity=0.95
            token_partitions=4
            level / speedlevel
            bt?
    '''
    support = supported_formats()

    profile, format = profile.split('.')
    bpp = 0.17
    video_codec = 'libvpx'
    audio_codec = 'libvorbis'

    if profile == '1080p':
        height = 1080

        audiorate = 48000
        audioquality = 6
        audiobitrate = None
        audiochannels = None
    elif profile == '720p':
        height = 720

        audiorate = 48000
        audioquality = 5
        audiobitrate = None
        audiochannels = None
    elif profile == '480p':
        height = 480

        audiorate = 44100
        audioquality = 3
        audiobitrate = None
        audiochannels = 2
    elif profile == '432p':
        height = 432
        audiorate = 44100
        audioquality = 2
        audiobitrate = None
        audiochannels = 2
    elif profile == '360p':
        height = 360

        audiorate = 44100
        audioquality = 1
        audiobitrate = None
        audiochannels = 1
    elif profile == '288p':
        height = 288

        audiorate = 44100
        audioquality = 0
        audiobitrate = None
        audiochannels = 1
    elif profile == '240p':
        height = 240

        audiorate = 44100
        audioquality = 0
        audiobitrate = None
        audiochannels = 1
    elif profile == '144p':
        height = 144

        audiorate = 22050
        audioquality = -1
        audiobitrate = '22k'
        audiochannels = 1
    else:
        height = 96

        #if support['vp9']:
        #    audio_codec = 'libopus'
        #    video_codec = 'libvpx-vp9'

        audiorate = 22050
        audioquality = -1
        audiobitrate = '22k'
        audiochannels = 1

    if format == 'webm' and audio_codec == 'libopus':
        audiorate = 48000
        if not audiobitrate:
            audiobitrate = '%sk' % {
                -1: 32,  0: 48,  1: 64,  2: 96,  3: 112,  4: 128,
                 5: 144, 6: 160, 7: 192, 8: 256, 9: 320, 10: 512,
            }[audioquality]
    if format == 'webm' and video_codec == 'libvpx-vp9':
        bpp = 0.15

    if info['video'] and 'display_aspect_ratio' in info['video'][0]:
        # dont make video bigger
        height = min(height, info['video'][0]['height'])

        dar = AspectRatio(info['video'][0]['display_aspect_ratio'])
        width = int(dar * height)
        width += width % 2

        aspect = dar.ratio
        # use 1:1 pixel aspect ratio if dar is close to that
        if abs(width/height - dar) < 0.02:
            aspect = '%s:%s' % (width, height)

        if info['video'][0]['framerate'] == '0:0':
            fps = 25
        else:
            fps = AspectRatio(info['video'][0]['framerate'])

        extra = []
        if fps == 50:
            fps = 25
            extra += ['-r', '25']
        if fps == 60:
            fps = 30
            extra += ['-r', '30']
        fps = min(float(fps), 30)

        bitrate = height*width*fps*bpp/1000

        video_settings = [
            '-vb', '%dk' % bitrate,
            '-aspect', aspect,
            '-g', '%d' % int(fps*5),
            '-vf', 'yadif,hqdn3d,scale=%s:%s' % (width, height),
        ] + extra
        if format == 'webm':
            video_settings += [
                '-vcodec', video_codec,
                '-quality', 'good',
                '-cpu-used', '1' if video_codec == 'libvpx-vp9' else '0',
                '-lag-in-frames', '16',
                '-auto-alt-ref', '1',
            ]
            if video_codec == 'libvpx-vp9':
                video_settings += [
                    '-tile-columns', '6',
                    '-frame-parallel', '1',
                ]
        if format == 'mp4':
            video_settings += [
                '-c:v', 'libx264',
                '-preset:v', 'medium',
                '-profile:v', 'high',
                '-level', '4.0',
                '-pix_fmt', 'yuv420p',
            ]
        video_settings += ['-map', '0:%s,0:0' % info['video'][0]['id']]
        audio_only = False
    else:
        video_settings = ['-vn']
        audio_only = True

    # ignore some unsupported audio codecs
    if info['audio'] and info['audio'][0].get('codec') in ('qdmc', ):
        audio_settings = ['-an']
    elif info['audio']:
        if video_settings == ['-vn'] or not info['video']:
            n = 0
        else:
            n = 1
        audio_settings = []
        # mix 2 mono channels into stereo(common for fcp dv mov files)
        audio_track = 0
        if audio_track == 0 and len(info['audio']) == 2 \
                and len(list(filter(None, [a['channels'] == 1 or None for a in info['audio']]))) == 2:
            audio_settings += [
                '-filter_complex',
                '[0:%s][0:%s] amerge' % (info['audio'][0]['id'], info['audio'][1]['id'])
            ]
            mono_mix = True
        else:
            mono_mix = False
            audio_settings += ['-map', '0:%s,0:%s' % (info['audio'][audio_track]['id'], n)]
        audio_settings += ['-ar', str(audiorate)]
        if audio_codec != 'libopus':
            audio_settings += ['-aq', str(audioquality)]
        if mono_mix:
            ac = 2
        else:
            ac = info['audio'][0].get('channels')
            if not ac:
                ac = audiochannels
        if audiochannels:
            ac = min(ac, audiochannels)
            audio_settings += ['-ac', str(ac)]
        if audiobitrate:
            audio_settings += ['-ab', audiobitrate]
        if format == 'mp4':
            audio_settings += ['-c:a', 'aac', '-strict', '-2']
        elif audio_codec == 'libopus':
            audio_settings += ['-c:a', 'libopus', '-frame_duration', '60']
        else:
            audio_settings += ['-c:a', audio_codec]
    else:
        audio_settings = ['-an']

    base = [command('ffmpeg'),
            #'-nostats', '-loglevel', 'error',
            '-y', '-i', video, '-threads', '4', '-map_metadata', '-1', '-sn']

    if format == 'webm':
        post = ['-f', 'webm', target]
    elif format == 'mp4':
        post = ['-movflags', '+faststart', '-f', 'mp4', target]
    else:
        post = [target]

    cmds = []
    if video_settings != ['-vn']:
        pass1_post = post[:]
        pass1_post[-1] = '/dev/null'
        if format == 'webm':
            pass1_post = ['-speed', '4'] + pass1_post
            post = ['-speed', '1'] + post
        cmds.append(base + ['-an', '-pass', '1', '-passlogfile', '%s.log' % target]
                         + video_settings + pass1_post)
        cmds.append(base + ['-pass', '2', '-passlogfile', '%s.log' % target]
                         + audio_settings + video_settings + post)
    else:
        cmds.append(base + audio_settings + video_settings + post)

    if not support.get(format):
        if format == 'webm':
            print("ffmpeg is compiled without WebM support")
        elif format == 'mp4':
            print("ffmpeg is compiled without H.264 support")
        else:
            print('trying to encode in an unsupported format')
        sys.exit(1)
    return cmds

def video(video, target, profile, info):
    enc_target = target + 'tmp.mp4' if target.endswith('.mp4') else target + '.tmp.webm'
    cmds = video_cmd(video, enc_target, profile, info)
    profile, format = profile.split('.')
    for cmd in cmds:
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             close_fds=True)
        try:
            p.wait()
            r = p.returncode
            print('Input:\t', video)
            print('Output:\t', target)
        except KeyboardInterrupt:
            p.kill()
            r = 1
            if os.path.exists(enc_target):
                print("\n\ncleanup unfinished encoding:\nremoving", enc_target)
                print("\n")
                os.unlink(enc_target)
            if os.path.exists(target):
                print("\n\ncleanup unfinished encoding:\nremoving", target)
                print("\n")
                os.unlink(target)
            if format == 'mp4' and os.path.exists("%s.mp4" % target):
                os.unlink("%s.mp4" % target)
            sys.exit(1)
        if p.returncode != 0:
            if os.path.exists(enc_target):
                os.unlink(enc_target)
            if os.path.exists(target):
                os.unlink(target)
            for f in glob('%s.log*' % enc_target):
                os.unlink(f)
            return False
        else:
            r = 0
    if r == 0 and enc_target != target:
        shutil.move(enc_target, target)
    for f in glob('%s.log*' % enc_target):
        os.unlink(f)
    return r == 0



#!/usr/bin/env python3
from setuptools import setup


def get_version():
    return 6 #import os
    import re
    import subprocess
    dot_git = os.path.join(os.path.dirname(__file__), '.git')
    changelog = os.path.join(os.path.dirname(__file__), 'debian/changelog')
    if os.path.exists(dot_git):
        cmd = ['git', 'rev-list', 'HEAD', '--count']
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        rev = int(stdout) - 267
        return u'%s' % rev
    elif os.path.exists(changelog):
        f = open(changelog)
        head = f.read().strip().split('\n')[0]
        f.close()
        rev = re.compile('\d+\.\d+\.(\d+)').findall(head)
        if rev:
            return u"%s" % rev[0]
    return u'unknown'


setup(
    name="pandora_client",
    version="0.3.%s" % get_version(),
    description="pandora_client is a commandline client for pan.do/ra. You can use it to import videos into a pan.do/ra system. It is currently known to work on Linux and Mac OS X.",
    author="j",
    author_email="j@mailb.org",
    url="http://wiki.0x2620.org/wiki/pandora_client",
    license="GPLv3",
    scripts=[
        'bin/pandora_client',
    ],
    packages=[
        'pandora_client'
    ],
    install_requires=[
        'ox >= 2.3.804,<3',
        'six',
        'requests >= 1.1.0',
        'zeroconf',
        'netifaces',
    ],
    keywords=[],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License (GPL)',
    ],
)
